from delphai_utils.logging import logging
from delphai_utils.config import get_config
from elasticsearch import AsyncElasticsearch

logging.getLogger("elasticsearch").setLevel(logging.ERROR)
es_host = get_config("elasticsearch.host")
es_index = get_config("elasticsearch.index")
es_username = get_config("elasticsearch.username")
es_password = get_config("elasticsearch.password")
es_port = get_config("elasticsearch.port")
es_address = es_host
if not es_address.startswith("https://"):
    es_address = f"http://{es_host}:{es_port or 80}"
es = AsyncElasticsearch(
    hosts=[es_address], http_compress=True, http_auth=(es_username, es_password)
)



from itertools import chain
import collections.abc
from typing import Iterable, Iterator, Optional, Union, cast
import asyncio
import socket
from grpc import Server, aio
from grpc_reflection.v1alpha import reflection
from grpc_health.v1 import health, health_pb2_grpc
from grpc_health.v1.health_pb2 import _HEALTH
from delphai_utils.logging import logging
from google.protobuf.descriptor import FileDescriptor, ServiceDescriptor
from delphai_utils.gateway import start_gateway
from delphai_utils.keycloak import update_public_keys
from prometheus_client import start_http_server
import functools
from grpc.aio import AioRpcError
from grpc import StatusCode
from delphai_utils.interceptors.metrics import MetricsInterceptor
from delphai_utils.interceptors.authentication import AuthenticationInterceptor

shutdown_event = asyncio.Event()


def grpc_error(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        context = args[2]
        try:
            result = await func(*args, **kwargs)
            return result
        except AioRpcError as ex:
            await context.abort(ex.code(), ex.details())
        except Exception as ex:
            await context.abort(StatusCode.INTERNAL, str(ex))

    return wrapper


def is_port_free(host, port):
    """
    determine whether `host` has the `port` free

    From: https://www.thepythoncode.com/article/make-port-scanner-python
    """
    s = socket.socket()
    try:
        s.connect((host, port))
    except Exception:
        return True
    else:
        return False


class NoPortFoundError(Exception):
    ...


def find_free_port(start: int, host="127.0.0.1", num_tries=4) -> int:
    for port in range(start, start + num_tries):
        if is_port_free(host, port):
            return port
        else:
            logging.info(f"Port {port} already in use.")
    message = f"No free port found in range [{start}, {start + num_tries - 1}]"
    logging.error(message)
    raise NoPortFoundError(message)


def create_grpc_server(
    descriptors: Union[FileDescriptor, Iterable[FileDescriptor]],
    server: Optional[aio.Server] = None,
) -> aio.Server:
    """Configures a grpc server based on one or multiple proto file descriptors.

    A existing grpc server can be passed as configuration base. If not the
    default configured grpc.aio.Server will be used.
    """
    if not isinstance(descriptors, collections.abc.Iterable):
        descriptors = [descriptors]
    max_message_length = 512 * 1024 * 1024
    server = server or aio.server(
        options=[
            ("grpc.max_send_message_length", max_message_length),
            ("grpc.max_receive_message_length", max_message_length),
        ],
        interceptors=(
            AuthenticationInterceptor(),
            MetricsInterceptor(),
        ),
    )

    health_servicer = health.HealthServicer(experimental_non_blocking=True)
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)

    services: Iterator[ServiceDescriptor] = chain.from_iterable(
        [descriptor.services_by_name.values() for descriptor in descriptors]
    )
    service_names = [service.full_name for service in services]
    all_service_names = (
        *service_names,
        _HEALTH.full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(all_service_names, server)
    server.__dict__["descriptors"] = descriptors
    return server


def start_server(
    server: Server,
    gateway: bool = True,
    grpc_port: Optional[int] = None,
    http_port: Optional[int] = None,
    metrics_port: Optional[int] = None,
):
    """Start a grpc server including gateway and background tasks."""
    logging.info("starting grpc server...")
    if not grpc_port:
        grpc_port = find_free_port(8080)
    server.add_insecure_port(f"[::]:{grpc_port}")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(server.start())
    logging.info(f"started grpc server on port {grpc_port}")
    if not metrics_port:
        metrics_port = find_free_port(9191)
    start_http_server(metrics_port)
    logging.info(f"started metrics server on port {metrics_port}")
    try:
        if gateway:
            if "descriptors" not in server.__dict__:
                raise RuntimeError(
                    "Server instance does not include the proto file descriptors. Make sure you instantiate it with 'create_grpc_server'"
                )
            if not http_port:
                http_port = find_free_port(7070)
            gateway_task = start_gateway(
                cast(
                    Union[FileDescriptor, Iterable[FileDescriptor]],
                    server.__dict__["descriptors"],
                ),
                grpc_port,
                http_port,
            )
            loop.create_task(gateway_task)
        loop.create_task(server.wait_for_termination())
        loop.create_task(update_public_keys())
        loop.run_forever()
    except KeyboardInterrupt:
        logging.info("stopped server (keyboard interrupt)")



import grpc
from jose.exceptions import JWTError
from delphai_utils.keycloak import decode_token, PublicKeyFetchError
from delphai_utils.logging import logging
from grpc.aio import Metadata, ServerInterceptor


class AuthenticationInterceptor(ServerInterceptor):
    error_message: str

    def __init__(self):
        async def abort(ignored_request, context):
            await context.abort(grpc.StatusCode.UNAUTHENTICATED, self.error_message)

        self._abort = grpc.unary_unary_rpc_method_handler(abort)

    async def intercept_service(self, continuation, handler_call_details):
        metadata: Metadata = handler_call_details.invocation_metadata
        authorization_header: str = None
        for metadatum in metadata:
            if metadatum[0] == "authorization":
                authorization_header = metadatum[1]
        if not authorization_header:
            logging.warning("authorization header not specified")
        else:
            if "Bearer " not in authorization_header:
                self.error_message = "Authorization header has the wrong format."
                logging.error(self.error_message, authorization_header)
                return self._abort
            _, access_token = authorization_header.split("Bearer ")
            try:
                await decode_token(access_token)
            except JWTError as ex:
                self.error_message = f"Error decoding the token: {ex}"
                logging.error(self.error_message)
                return self._abort
            except PublicKeyFetchError as ex:
                self.error_message = f"Error fetching jwk from keycloak: {ex}"
                logging.error(self.error_message)
                return self._abort
        return await continuation(handler_call_details)



from typing import Dict, List, Tuple, Union
from grpc.aio import ServicerContext
from delphai_utils.keycloak import decode_token
import asyncio
import nest_asyncio
import json
from base64 import b64decode

nest_asyncio.apply()


def get_user(context: ServicerContext) -> Dict:
    """
    Get x-delphai-user information.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :raises: KeyError
    :return x-delphai-user dictionary.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: Dict
    """

    metadata = dict(context.invocation_metadata())
    if "authorization" in metadata:
        authorization_header = metadata["authorization"]
        _, access_token = authorization_header.split("Bearer ")
        decoded_access_token = asyncio.get_event_loop().run_until_complete(
            decode_token(access_token)
        )
        user = {
            "https://delphai.com/mongo_user_id": decoded_access_token.get(
                "mongo_user_id"
            ),
            "https://delphai.com/client_id": decoded_access_token.get(
                "mongo_client_id"
            ),
        }
        if (
            "realm_access" in decoded_access_token
            and "roles" in decoded_access_token.get("realm_access")
        ):
            roles = decoded_access_token.get("realm_access").get("roles")
            user["roles"] = roles
        if "group_membership" in decoded_access_token:
            user["groups"] = decoded_access_token["group_membership"]
        if "limited_dataset_group_name" in decoded_access_token:
            user["limited_dataset_group_name"] = decoded_access_token[
                "limited_dataset_group_name"
            ]
        if "name" in decoded_access_token:
            user["name"] = decoded_access_token["name"]
        return user
    else:
        user64 = metadata["x-delphai-user"]
        user = b64decode(user64)
        return json.loads(user)


def get_groups(context: ServicerContext) -> List[str]:
    """
    Gets groups of calling identity
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return raw roles passed from keycloak.
      For example this can be ["/delphai/Development"]
    :rtype List[str]
    """

    assert isinstance(context, object)
    try:
        user = get_user(context)
    except KeyError:
        return []
    return user.get("groups") or []


def get_affiliation(
    context: ServicerContext,
) -> Tuple[Union[str, None], Union[str, None]]:
    """
    Gets organization and department of user
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return organization and department as a tuple or None if not affiliated
    :rtype Union[Tuple[str, None], Tuple[str, None]]
    """

    raw_groups = get_groups(context)
    if len(raw_groups) == 0:
        return None, None
    else:
        group_hirarchy = raw_groups[0].split("/")[1:3]
        if len(group_hirarchy) == 1:
            return group_hirarchy.pop(), None
        elif len(group_hirarchy) == 2:
            return tuple(group_hirarchy)
        else:
            return None, None


def get_user_and_client_ids(context: ServicerContext) -> Tuple[str, str]:
    """
    Get user_id and client_id.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return user_id and client_id.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: Tuple[str, str]
    """
    try:
        user = get_user(context)
        user_id = user.get("https://delphai.com/mongo_user_id")
        client_id = user.get("https://delphai.com/client_id")
    except Exception:
        return "", ""
    return user_id, client_id



from logging import error, warn, info, debug  # noqa: F401
import logging
import logging.config
from sys import exc_info
from traceback import format_tb
from deepmerge import always_merger
import coloredlogs
from delphai_utils.config import get_config

default_config = {
    "version": 1,
    "disable_existing_loggers": True,
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        }
    },
    "formatters": {
        "standard": {
            "format": "[%(asctime)s] [%(levelname)s] %(message)s",
            "datefmt": "%H:%M:%S",
        }
    },
    "loggers": {"": {"level": "INFO", "handlers": ["default"]}},
}


def _configure_logging():
    provided_config = {}
    try:
        provided_config = get_config("logging")
    except Exception:
        pass
    logging_config = always_merger.merge(default_config, provided_config)
    standard_format = logging_config["formatters"]["standard"]

    logging.config.dictConfig(logging_config)
    level = logging.getLevelName(logging.getLogger().getEffectiveLevel())
    coloredlogs.install(
        level=level, fmt=standard_format["format"], datefmt=standard_format["datefmt"]
    )


def error_with_traceback():
    """
    Writes to log an occured exception with a last line of traceback
    """
    einfo = exc_info()
    exc_class = einfo[0].__name__
    exc_descr = str(einfo[1])
    traceback = format_tb(einfo[2])
    traceback = "\n" + traceback[-1].strip() if traceback else ""
    logging.error(f"{exc_class}: {exc_descr}{traceback}")


_configure_logging()



import asyncio
from logging import Logger
from aioprometheus.collectors import Histogram
from deepmerge import always_merger
from delphai_utils.logging import default_config as default_logging_config, logging
from abc import abstractmethod, abstractproperty
import faust
import socket
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, TypeVar, Type
from delphai_utils.config import get_config
from faust.types import CodecArg
from faust.types.models import ModelT
from faust.streams import Stream
from confluent_kafka.admin import AdminClient, NewTopic
from betterproto import Message
from faust.serializers import Codec
from dacite import from_dict
from delphai_utils.config import get_config
import coloredlogs
from aioprometheus import Service, Counter, Summary
from timeit import default_timer


logging.getLogger("aiohttp").setLevel(logging.ERROR)


def is_port_free(host, port):
    """
    determine whether `host` has the `port` free

    From: https://www.thepythoncode.com/article/make-port-scanner-python
    """
    s = socket.socket()
    try:
        s.connect((host, port))
    except Exception:
        return True
    else:
        return False


class NoPortFoundError(Exception):
    ...


def find_free_port(start: int, host="127.0.0.1", num_tries=4) -> int:
    for port in range(start, start + num_tries):
        if is_port_free(host, port):
            return port
        else:
            logging.info(f"Port {port} already in use.")
    message = f"No free port found in range [{start}, {start + num_tries - 1}]"
    logging.error(message)
    raise NoPortFoundError(message)


@dataclass
class Step:
    name: str
    partitions: int
    output: Optional[str]
    tables: Optional[List[str]] = field(default_factory=lambda: [])


T = TypeVar("T")


class FaustProtobufSerializer(Codec):
    def __init__(self, type_class: T, **kwargs):
        self.type_class = type_class
        super(FaustProtobufSerializer, self).__init__(type_class=type_class, **kwargs)

    def _loads(self, s: bytes) -> T:
        return self.type_class().from_json(s.decode("utf-8"))

    def _dumps(self, s: Message) -> bytes:
        return s.to_json().encode("utf-8")


@dataclass
class Metrics:
    inbound_counter: Optional[Counter] = None
    processed_counter: Optional[Counter] = None
    lag_timer: Optional[Summary] = None


class FaustAgent:
    step: str
    step_config: Step
    agent_id: str
    input_topic: faust.TopicT
    output_topic: faust.TopicT
    input_value_type: Type[ModelT]
    output_value_type: Type[ModelT]
    input_value_serializer: CodecArg
    output_value_serializer: CodecArg
    agent: faust.Agent
    logger: Logger
    metrics: Metrics
    metrics_server: Service

    def setup_metrics(self):
        self.metrics = Metrics()
        name = self.agent_id.replace(".", "_").replace("-", "_")
        self.metrics.inbound_counter = Counter(
            f"delphai_{name}_inbound_counter",
            "incoming requests",
            const_labels={
                "step": self.step,
                "pipeline": self.agent.app.conf.id,
            },
        )
        self.metrics.processed_counter = Counter(
            f"delphai_{name}_processed_counter",
            "processed requests",
            const_labels={
                "step": self.step,
                "pipeline": self.agent.app.conf.id,
            },
        )
        self.metrics.lag_timer = Histogram(
            f"delphai_{name}_lag_timer",
            "time to process requests",
            const_labels={
                "step": self.step,
                "pipeline": self.agent.app.conf.id,
            },
        )
        self.metrics_server.register(self.metrics.inbound_counter)
        self.metrics_server.register(self.metrics.processed_counter)
        self.metrics_server.register(self.metrics.lag_timer)

    @abstractproperty
    def step(self) -> str:
        pass

    @abstractproperty
    def input_value_type(self) -> Type[ModelT]:
        pass

    @abstractproperty
    def output_value_type(self) -> Type[ModelT]:
        pass

    @abstractproperty
    def input_value_serializer(self) -> CodecArg:
        pass

    @abstractproperty
    def output_value_serializer(self) -> CodecArg:
        pass

    @abstractmethod
    async def process(self, requests: Stream[Type[ModelT]]):
        pass

    async def on_start(self):
        self.logger.info(f"started agent {self.agent_id}")

    async def after_attach(self):
        pass

    def start_measure(self, labels: Dict[str, str] = {}):
        start = default_timer()
        self.metrics.inbound_counter.inc(labels)
        return start

    def end_measure(self, start: float, labels: Dict[str, str] = {}):
        end = default_timer()
        self.metrics.lag_timer.add(labels, end - start)
        self.metrics.processed_counter.inc(labels)


class FaustApp:
    app: faust.App
    worker: faust.Worker
    logging_config: Dict
    faust_config: Dict
    metrics: Service = Service()

    async def on_startup_finished(self):
        self.logger.info(f"started app {self.app._conf.id}")

    async def on_start(self):
        self.logger.info(f"starting app {self.app._conf.id}")

    async def on_shutdown(self):
        self.logger.info(f"shutting down {self.app._conf.id}")

    def on_setup_root_logger(self, logger: logging.Logger, level: int):
        standard_format = self.logging_config["formatters"]["standard"]
        effective_level = self.logger.getEffectiveLevel()
        logger.setLevel(effective_level)
        coloredlogs.install(
            level=effective_level,
            fmt=standard_format["format"],
            datefmt=standard_format["datefmt"],
            logger=logger,
        )
        self.worker._disable_spinner_if_level_below_WARN(effective_level)
        self.worker._setup_spinner_handler(logger, effective_level)

    def __init__(self, id: str, broker: str) -> None:
        try:
            self.faust_config = get_config("faust")
            if self.faust_config is None:
                self.faust_config = {}
        except Exception:
            pass
        self.app = faust.App(
            id=id, broker=broker, web_enabled=False, **self.faust_config
        )
        loop = asyncio.get_event_loop()
        provided_logging_config = {}
        try:
            provided_logging_config = get_config("logging")
        except Exception:
            pass
        default_logging_config["disable_existing_loggers"] = False
        default_logging_config["formatters"]["standard"][
            "format"
        ] = "[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s"
        default_logging_config["loggers"]["faust"] = {"level": logging.ERROR}
        default_logging_config["loggers"]["aiokafka"] = {"level": logging.ERROR}
        default_logging_config["loggers"]["mode"] = {"level": logging.ERROR}
        self.logging_config = always_merger.merge(
            default_logging_config, provided_logging_config
        )
        self.worker = faust.Worker(
            self.app,
            loop=loop,
            loglevel=logging.INFO,
            logging_config=self.logging_config,
        )
        self.worker.on_startup_finished = self.on_startup_finished
        self.worker.on_start = self.on_start
        self.worker.on_shutdown = self.on_shutdown
        self.worker.on_setup_root_logger = self.on_setup_root_logger
        self.logger = logging.getLogger(self.app._conf.id)

    def start(self, agents: List[FaustAgent]):
        try:
            broker = str(self.app._conf.broker[0])
            steps = list(map(lambda s: s["name"], get_config("steps")))
            for step in steps:
                self.create_topic(broker, step)
            tasks = []
            for agent in agents:
                tasks.append(self.attach(agent))
            metrics_port = find_free_port(9191)
            asyncio.get_event_loop().create_task(self.metrics.start(port=metrics_port))
            self.logger.info(f"started metrics server on port {metrics_port}")
            self.worker.loop.run_until_complete(asyncio.gather(*tasks))
            self.worker.execute_from_commandline()
        except KeyboardInterrupt:
            self.logger.info("keyboard interrupt received")

    def get_step_config(self, step: str) -> Step:
        step_config = next((s for s in get_config("steps") if s["name"] == step), None)
        return from_dict(data_class=Step, data=step_config)

    def create_topic(self, broker: str, step: str):
        step_config = self.get_step_config(step)
        client = AdminClient({"bootstrap.servers": broker.replace("kafka://", "")})
        topics = list(client.list_topics().topics.keys())
        topic_names = [f"{self.app._conf.id}.{step_config.name}"]
        for table in step_config.tables:
            table_topic_name = f"{self.app._conf.id}-{table}-changelog"
            topic_names.append(table_topic_name)
        for topic_name in topic_names:
            if topic_name not in topics:
                self.logger.info(
                    f"creating topic {topic_name} with {step_config.partitions} partitions"
                )
                resp = client.create_topics(
                    [NewTopic(topic_name, step_config.partitions, 1)]
                )
                resp[topic_name].result()
            else:
                self.logger.info(f"topic {topic_name} already exists")

    async def attach(self, agent: FaustAgent):
        agent.step_config = self.get_step_config(agent.step)
        agent.agent_id = f"{self.app._conf.id}.{agent.step}"
        agent.logger = logging.getLogger(f"{self.app._conf.id}] [{agent.step}")
        agent.input_topic = self.app.topic(
            agent.agent_id,
            value_type=agent.input_value_type,
            value_serializer=agent.input_value_serializer,
        )
        output_topic_name = f"{self.app._conf.id}.{agent.step_config.output}"
        agent.output_topic = self.app.topic(
            output_topic_name,
            value_serializer=agent.output_value_serializer,
        )
        new_agent = self.app.agent(agent.input_topic)(agent.process)
        agent.agent = new_agent
        agent.metrics_server = self.metrics
        agent.setup_metrics()
        await agent.after_attach()
        await new_agent.start()



import collections.abc
from typing import Dict, Iterable, List, Union
from google.protobuf.descriptor import (
    FieldDescriptor,
    FileDescriptor,
    MethodDescriptor,
    ServiceDescriptor,
)
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
from grpc import Channel, StatusCode
from delphai_utils.logging import logging
from google.protobuf.descriptor_pb2 import MethodOptions
from google.api.http_pb2 import HttpRule
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.exceptions import HTTPException
from starlette.middleware.base import BaseHTTPMiddleware
from grpc.experimental.aio import insecure_channel, AioRpcError
from starlette.middleware.cors import CORSMiddleware
import os
from google.protobuf import symbol_database
from hypercorn.config import Config
from hypercorn.asyncio import serve
from urllib.parse import urlparse
from time import perf_counter
import json
from base64 import b64encode
from starlette_prometheus import metrics, PrometheusMiddleware
from delphai_utils.utils import generate_default_message
from jose.exceptions import JWTError
from delphai_utils.keycloak import decode_token, PublicKeyFetchError

supported_methods = ["get", "put", "post", "delete", "patch"]
status_map = {
    StatusCode.OK: 200,
    StatusCode.CANCELLED: 499,
    StatusCode.UNKNOWN: 500,
    StatusCode.INVALID_ARGUMENT: 400,
    StatusCode.DEADLINE_EXCEEDED: 504,
    StatusCode.NOT_FOUND: 404,
    StatusCode.ALREADY_EXISTS: 409,
    StatusCode.PERMISSION_DENIED: 403,
    StatusCode.UNAUTHENTICATED: 401,
    StatusCode.RESOURCE_EXHAUSTED: 429,
    StatusCode.FAILED_PRECONDITION: 412,
    StatusCode.ABORTED: 499,
    StatusCode.OUT_OF_RANGE: 416,
    StatusCode.UNIMPLEMENTED: 501,
    StatusCode.INTERNAL: 500,
    StatusCode.UNAVAILABLE: 503,
    StatusCode.DATA_LOSS: 420,
}


class AccessLogMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        start = perf_counter()
        response = await call_next(request)
        if response.status_code < 400:
            path = urlparse(str(request.url)).path
            end = perf_counter()
            elapsed = round((end - start) * 1000, 2)
            logging.info(
                f"[{response.status_code}] {request.method} {path} [{elapsed}ms]"
            )
        return response


async def process_grpc_request(
    descriptor: FileDescriptor,
    service_name: str,
    method: str,
    input: Message,
    channel: Channel,
    metadata: Dict = {},
):
    service = descriptor.services_by_name[service_name]
    method_descriptor = service.methods_by_name[method]
    input_prototype = symbol_database.Default().GetPrototype(
        method_descriptor.input_type
    )
    output_prototype = symbol_database.Default().GetPrototype(
        method_descriptor.output_type
    )
    method_callable = channel.unary_unary(
        f"/{service.full_name}/{method}",
        request_serializer=input_prototype.SerializeToString,
        response_deserializer=output_prototype.FromString,
    )
    response = await method_callable(
        input_prototype(**input), metadata=metadata.items()
    )
    return response


async def http_exception(request, exc):
    if "favicon.ico" in str(request.url):
        detail = "not found"
        status_code = 404
    else:
        path = urlparse(str(request.url)).path
        logging.error(f"[{exc.status_code}] {request.method} {path} - {exc.detail}")
        detail = exc.detail
        status_code = exc.status_code
    return JSONResponse(
        {"detail": detail, "status": exc.status_code}, status_code=status_code
    )


def get_http_handlers(
    descriptor: FileDescriptor, service_name: str, method: str, channel: Channel
):
    async def method_get_handler(request):
        service = descriptor.services_by_name[service_name]
        method_descriptor = service.methods_by_name[method]
        input = generate_default_message(method_descriptor.input_type)
        output = generate_default_message(method_descriptor.output_type)

        function_name = f"{service.full_name}.{method}"
        return JSONResponse(
            {"function_name": function_name, "input": input, "output": output}
        )

    async def request_handler(request: Request):
        body = {}
        if len(await request.body()) > 0:
            body = await request.json()
        input = {**request.path_params, **request.query_params, **body}
        metadata = {}
        if "Authorization" not in request.headers:
            logging.warn("no authorization header")
        else:
            authorization_header = request.headers.get("Authorization")
            if "Bearer " not in authorization_header:
                error_msg = "Authorization header has the wrong format."
                logging.error(error_msg, authorization_header)
                raise HTTPException(401, detail=error_msg)
            _, access_token = authorization_header.split("Bearer ")
            try:
                decoded_access_token = await decode_token(access_token)
            except JWTError as ex:
                error_msg = f"Error decoding the token: {ex}"
                logging.error(error_msg)
                raise HTTPException(401, detail=error_msg)
            except PublicKeyFetchError as ex:
                error_msg = f"Error fetching jwk from keycloak: {ex}"
                logging.error(error_msg)
                raise HTTPException(502, detail=error_msg)
            user = {
                "https://delphai.com/mongo_user_id": decoded_access_token.get(
                    "mongo_user_id"
                ),
                "https://delphai.com/client_id": decoded_access_token.get(
                    "mongo_client_id"
                ),
            }
            if (
                "realm_access" in decoded_access_token
                and "roles" in decoded_access_token.get("realm_access")
            ):
                roles = decoded_access_token.get("realm_access").get("roles")
                user["roles"] = roles
            if "group_membership" in decoded_access_token:
                user["groups"] = decoded_access_token["group_membership"]
            if "limited_dataset_group_name" in decoded_access_token:
                user["limited_dataset_group_name"] = decoded_access_token[
                    "limited_dataset_group_name"
                ]
            if "name" in decoded_access_token:
                user["name"] = decoded_access_token["name"]
            user_json = json.dumps(user).encode("ascii")
            user_b64 = b64encode(user_json).decode("utf-8")
            metadata = {"x-delphai-user": user_b64}

        try:
            raw_output = await process_grpc_request(
                descriptor, service_name, method, input, channel, metadata=metadata
            )
            output = MessageToDict(raw_output, preserving_proto_field_name=True)
            return JSONResponse(output)
        except AioRpcError as ex:
            detail = ex.details()
            grpc_status = ex.code()
            http_status_code = status_map[grpc_status]
            raise HTTPException(http_status_code, detail=detail)
        except Exception as ex:
            detail = str(ex).replace("\n", " ")
            http_status_code = 500
            raise HTTPException(http_status_code, detail=detail)

    return method_get_handler, request_handler


async def start_gateway(
    descriptors: Union[FileDescriptor, Iterable[FileDescriptor]],
    grpc_port: int,
    http_port: int = 7070,
):
    if not isinstance(descriptors, collections.abc.Iterable):
        descriptors = [descriptors]
    try:
        logging.info("starting gateway...")
        debug = (
            "DELPHAI_ENVIRONMENT" in os.environ
            and os.environ["DELPHAI_ENVIRONMENT"] == "development"
        )
        app = Starlette(debug=debug)
        app.add_exception_handler(HTTPException, http_exception)
        app.add_middleware(AccessLogMiddleware)
        app.add_middleware(PrometheusMiddleware)
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
            allow_credentials=True,
        )
        app.add_exception_handler(HTTPException, http_exception)
        app.add_route("/metrics/", metrics)
        max_length = 512 * 1024 * 1024
        client_options = [
            ("grpc.max_send_message_length", max_length),
            ("grpc.max_receive_message_length", max_length),
        ]
        channel = insecure_channel(f"localhost:{grpc_port}", options=client_options)
        config = Config()
        config.bind = [f"0.0.0.0:{http_port}"]
        for descriptor in descriptors:
            for service_name in descriptor.services_by_name:
                service_handler: ServiceDescriptor = descriptor.services_by_name[
                    service_name
                ]
                if service_handler.full_name.startswith("grpc."):
                    logging.info(f"skipping service {service_handler.name}")
                else:
                    logging.info(f"processing service {service_handler.name}")
                    for method_name in service_handler.methods_by_name.keys():
                        method_get_handler, request_handler = get_http_handlers(
                            descriptor, service_handler.name, method_name, channel
                        )
                        service_name = method_name[1:].split("/")[0].split(".")[-1]
                        app.add_route(
                            f"/{service_handler.full_name}.{method_name}",
                            route=method_get_handler,
                            methods=["get"],
                        )
                        app.add_route(
                            f"/{service_handler.full_name}.{method_name}",
                            route=request_handler,
                            methods=["post"],
                        )
                        logging.info(f"  processing {method_name}")
                        method_descriptor: MethodDescriptor = (
                            service_handler.methods_by_name[method_name]
                        )
                        method_options: MethodOptions = method_descriptor.GetOptions()
                        fields: List[FieldDescriptor] = method_options.ListFields()
                        for field in fields:
                            if field[0].full_name == "google.api.http":
                                http_rule: HttpRule = field[1]
                                for supported_method in supported_methods:
                                    http_path = getattr(http_rule, supported_method)
                                    if http_path != "":
                                        app.add_route(
                                            http_path,
                                            route=request_handler,
                                            methods=[supported_method],
                                        )
        logging.info(f"started gateway on port {http_port}")
        return await serve(app, config)
    except Exception as ex:
        logging.error(str(ex))
